/* $Id: trio64.c,v 1.11 1998/11/23 20:37:12 marcus Exp $
******************************************************************************

   S3 86c765 (Trio64V+) integrated RAMDAC driver

   Copyright (C) 1997 Steffen Seeger	[seeger@ggi-project.org]
   Copyright (C) 1997 Jon M. Taylor	[taylorj@csus.edu]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Jon_Taylor
#define	DRIVER_NAME	"S3 Trio64V+ integrated RAMDAC"
#define	DRIVER_REV	"$Revision: 1.11 $"

#define	DEBUG_LEVEL	0
#define DAC_PRECISION	8

#include <kgi/module.h>
#include <kgi/types.h>
#include <kgi/kgi_commands.h>
#include <S3/Trio.h>
#include <S3/sdac.h>
#define	MODES (GM_TEXT16|GM_1BIT|GM_4BIT|GM_8BIT|GM_15BIT|GM_16BIT|GM_32BIT)

static struct kgi_ramdac ramdac = {
	"S3 Incorporated",			/* Manufacturer		*/
	"Trio64",				/* Model		*/
	GM_TEXT16 | GM_1BIT | GM_4BIT |		/* Supported modes	*/
	GM_8BIT | GM_15BIT | GM_16BIT |
	/* GM_24BIT | */ GM_32BIT,
	FL_DAC_8BIT,				/* Flags		*/
	{ 0x0FFF, 0x0FFF },			/* Maximum resolution	*/
	NULL,					/* Private data hook	*/
	DAC_8BIT_PB,				/* Device type		*/
	{ 0,  80000000 },			/* Normal mode		*/
	{ 0,         0 },			/* 2xclk mode		*/
	{
		{ 0,  75000000 },		/* KGIGT_TEXT		*/
		{ 0,         0 },		/* KGIGT_TEXT32		*/
		{ 0,  80000000 },		/* KGIGT_1BIT		*/
		{ 0,  80000000 },		/* KGIGT_4BIT		*/
		{ 0,  80000000 },		/* KGIGT_8BIT		*/	
		{ 0,  80000000 },		/* KGIGT_15BIT		*/
		{ 0,  80000000 },		/* KGIGT_16BIT		*/
		{ 0,         0 },		/* KGIGT_24BIT		*/
		{ 0,  75000000 }		/* KGIGT_32BIT		*/
	}
};

/* Driver state info 
*/
struct dac_private 
{
  kgi_uint	cur_hotx, cur_hoty;		/* Cursor hot spot	*/
  kgi_uint	cur_x, cur_y;			/* Current position	*/
  kgiu8	ctrl0, ctrl1, ctrl2, ctrl3;	/* Control registers	*/
};

#define	CTRL0		private->ctrl0
#define	CTRL1		private->ctrl1
#define	CTRL2		private->ctrl2
#define	CTRL3		private->ctrl3
#define	CUR_HOTX	private->cur_hotx
#define	CUR_HOTY	private->cur_hoty
#define	CUR_X		private->cur_x
#define	CUR_Y		private->cur_y
#define	GET_PRIVATE	struct dac_private *private = KGIM_PRIVATE(dpy)->ramdac

#include <IBM/clut.inc>

/* KGI interface 
*/
int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	
	struct dac_private *private;
	if (!(private = kgi_malloc_small(sizeof(*private)))) 
	{
		
		NOTICE1("failed to allocate private data!");
		return -ENOMEM;
	}
	
	KGIM_PRIVATE(dpy, ramdac) = private;
	memset(private, 0, sizeof(*private));
	card->ramdac = &ramdac;
	
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	kgi_free_small(KGIM_PRIVATE(dpy, ramdac));
	KGIM_PRIVATE(dpy, ramdac) = NULL;
}

int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode, 
			    enum kgi_tm_cmd *cmd)
{
	int div;
	int newclock = mode->dclk;
	struct kgi_ramdac *dac = dpy->card->ramdac;
	
	switch(mode->request.graphtype) 
	{
		
	case KGIGT_TEXT16:	
		div = 1;
		break;
		
	case KGIGT_1BIT:
		div = 32;
		break;
		
	case KGIGT_4BIT:
		div = 8;
		break;
		
	case KGIGT_8BIT:
		div = 4;
		break;
		
	case KGIGT_15BIT:
	case KGIGT_16BIT:	
		div = 2;
		break;
		
	case KGIGT_24BIT:
		return -E(RAMDAC, UNKNOWN);
		
	case KGIGT_32BIT:	
		div = 1;
		break;
		
	default:
		DEBUG1("Could not handle mode %x", mode->request.graphtype);
		return -E(RAMDAC, UNKNOWN);
	}

	switch (*cmd) 
	{
		
	case CMD_PROPOSE:
		mode->lclk.mul = 1;
		mode->clk.mul = 1; 
		mode->clk.div = 1;
		mode->lclk.div = div;
		DEBUG1("%i:%i multiplex mode", mode->lclk.div, mode->lclk.mul);
		mode->flags &= ~FL_DAC_2XCLOCK;
		mode->flags |= FL_DAC_8BIT;
		if ((mode->request.graphtype == KGIGT_32BIT) || 
		    (mode->request.graphtype == KGIGT_15BIT) ||
		    (mode->request.graphtype == KGIGT_16BIT)) 
		{
			
			mode->flags |= FL_DAC_GAMMA;
			
		}
		
		return EOK;
		
	case CMD_LOWER:
		if ((newclock < dac->clock[mode->request.graphtype].min) ||
		    (newclock < dac->normal.min)) 
		{
			
			DEBUG1("%i Hz DCLK is too low", newclock);
			return -E(RAMDAC, UNKNOWN);
			
		}
		
		if (dac->clock[mode->request.graphtype].max < newclock) 
		{
			
			newclock = dac->clock[mode->request.graphtype].max;
		}
		
		if (dac->normal.max < newclock) 
		{
			
			NOTICE1("Double clock mode not yet implemented");
			newclock = dac->normal.max;
			
		}
		
		DEBUG1("DCLK = %i", newclock);
		mode->dclk = newclock;
		return EOK;
		
	case CMD_RAISE:
		if ((newclock > dac->clock[mode->request.graphtype].max) ||
		    (newclock > dac->normal.max)) 
		{
			
			DEBUG1("%i Hz DCLK is too high", newclock);
			return -E(RAMDAC, UNKNOWN);
			
		}
		
		if (newclock < dac->clock[mode->request.graphtype].min) 
		{
			
			newclock = dac->clock[mode->request.graphtype].min;
			
		}
		
		if (newclock < dac->normal.min) 
		{
			
			newclock = dac->normal.min;
		}
		
		mode->dclk = newclock;
		return EOK;
		
	case CMD_CHECK:
		
		if ((mode->lclk.mul != 1) || (mode->lclk.div != div) ||
		    (mode->clk.mul != 1) || (mode->clk.div != 1)) 
		{
			
			return -E(RAMDAC, UNKNOWN);
		}
		
		if ((mode->dclk < dac->normal.min) ||
		    (mode->dclk < dac->clock[mode->request.graphtype].min))
		{
			
			return -E(RAMDAC, UNKNOWN);
		}
		
		if ((mode->dclk > dac->normal.max) ||
		    (mode->dclk > dac->clock[mode->request.graphtype].max))
		{
			
			return -E(RAMDAC, UNKNOWN);
		}
		
		return EOK;
		
	default:
		INTERNAL_ERROR;
		return -E(RAMDAC, UNKNOWN);
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	switch (mode->request.graphtype) 
	{
		
	case KGIGT_1BIT:
	case KGIGT_TEXT16:
	case KGIGT_4BIT:
	case KGIGT_8BIT:
	case KGIGT_15BIT:
	case KGIGT_16BIT:
	case KGIGT_24BIT:
	case KGIGT_32BIT:
		break;
		
	default:
		INTERNAL_ERROR;
	}

	dpy->set_clut = dac_set_clut;
}

int kgim_ramdac_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg) 
{
	/* This may set:
	 * - Hardware cursor
	 * - CLUT (or gamma correction table)
	 */
	
	kgi_clut clut;

	switch (cmd) 
	{ 
		
	case RAMDAC_GETINFO:
		graph->io_res = &ramdac;
		return EOK;
		
	case RAMDAC_SETINFO:
		if (suser()) 
		{
			
			memcpy(&ramdac, graph->io_buf, sizeof(ramdac));
			return EOK;
		} 
		else 
		{
			
			return -EPERM;
		}
		
	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data = (kgi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy, &clut, 0, 256);
		return EOK;
		
	case RAMDAC_GETCLUT:
		clut.size = 256;
		clut.data = (kgi_color *)(graph->io_buf);
		graph->io_res = graph->io_buf;
		dac_get_clut(graph->dev.dpy, &clut, 0, 256);
		return EOK;
	}
	return -E(RAMDAC,NOSUP_ALWAYS_CANT);  /* The driver can't do this */
}
